/* These are the uw-copyright.h and usc-copyright.h files that appear in
   the Prospero distribution.  */

/* 
  Copyright (c) 1989, 1990, 1991 by the University of Washington

  Permission to use, copy, modify, and distribute this software and its
  documentation for non-commercial purposes and without fee is hereby
  granted, provided that the above copyright notice appear in all copies
  and that both the copyright notice and this permission notice appear in
  supporting documentation, and that the name of the University of
  Washington not be used in advertising or publicity pertaining to
  distribution of the software without specific, written prior
  permission.  The University of Washington makes no representations
  about the suitability of this software for any purpose.  It is
  provided "as is" without express or implied warranty.

  Prospero was written by Clifford Neuman (bcn@isi.edu).

  Questions concerning this software should be directed to 
  info-prospero@isi.edu.

  */

/* 
  Copyright (c) 1991, 1992 by the University of Southern California
  All rights reserved.

  Permission to use, copy, modify, and distribute this software and its
  documentation in source and binary forms for non-commercial purposes
  and without fee is hereby granted, provided that the above copyright
  notice appear in all copies and that both the copyright notice and
  this permission notice appear in supporting documentation. and that
  any documentation, advertising materials, and other materials related
  to such distribution and use acknowledge that the software was
  developed by the University of Southern California, Information
  Sciences Institute.  The name of the University may not be used to
  endorse or promote products derived from this software without
  specific prior written permission.

  THE UNIVERSITY OF SOUTHERN CALIFORNIA makes no representations about
  the suitability of this software for any purpose.  THIS SOFTWARE IS
  PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

  Other copyrights might apply to parts of the Prospero distribution
  and are so noted when applicable.

  Prospero was originally written by Clifford Neuman (bcn@isi.edu).
  Contributors include Benjamin Britt (britt@isi.edu)
  and others identified in individual modules.

  Questions concerning this software should be directed to 
  info-prospero@isi.edu.

  */
